@@ -27,13 +27,12 @@
 #include <sys/mount.h>
 
 #include "global.h"
-#include <linux/gfs_ondisk.h>
 #include "copyright.cf"
 
 #define EXTERN
 #include "hexedit.h"
 #include "gfshex.h"
-#include "linux/gfs_ondisk.h"
+#include "gfs_ondisk.h"
 #include "linux_endian.h"
 
 #include <syslog.h>
@@ -141,9 +140,12 @@ int bobgets(char string[],int x,int y,int sz)
 			case(KEY_DC):
 			case(0x07F):
 				if (runningy>=y) {
-					char *p=&string[runningy-y];
-					while (*p)
-						*p++=*(p+1);
+					char *p;
+					p = &string[runningy-y];
+					while (*p) {
+						*p = *(p+1);
+						p++;
+					}
 					*p='\0';
 					runningy--;
 					// remove the character from the string 
@@ -157,9 +159,13 @@ int bobgets(char string[],int x,int y,int sz)
 				break;
 			case(KEY_BACKSPACE):
 				if (runningy>y) {
-					char *p=&string[runningy-y-1];
-					while (*p)
-						*p++=*(p+1);
+					char *p;
+
+					p = &string[runningy-y-1];
+					while (*p) {
+						*p = *(p+1);
+						p++;
+					}
 					*p='\0';
 					runningy--;
 					// remove the character from the string 
@@ -271,7 +277,7 @@ void print_usage(void)
 /* returns: metatype if block is a GFS structure block type                 */
 /*          0 if block is not a GFS structure                               */
 /* ------------------------------------------------------------------------ */
-int display_block_type(const unsigned char *lpBuffer)
+int display_block_type(const char *lpBuffer)
 {
 	int ret_type = 0; /* return type */
 
@@ -374,7 +380,7 @@ int display_block_type(const unsigned char *lpBuffer)
 /* ------------------------------------------------------------------------ */
 /* hexdump - hex dump the filesystem block to the screen                    */
 /* ------------------------------------------------------------------------ */
-int hexdump(uint64 startaddr, const unsigned char *lpBuffer, int len)
+int hexdump(uint64 startaddr, const char *lpBuffer, int len)
 {
 	const unsigned char *pointer,*ptr2;
 	int i;
@@ -540,7 +546,7 @@ int main(int argc, char *argv[])
 		if (!strcasecmp(argv[i], "-verbose"))
 			verbose = TRUE;
 		else if (!strcasecmp(argv[i], "-V")) {
-			printf("%s %s (built %s %s)\n", prog_name, GFS_RELEASE_NAME,
+			printf("%s %s (built %s %s)\n", prog_name, RELEASE_VERSION,
 				__DATE__, __TIME__);
 			printf("%s\n", REDHAT_COPYRIGHT);
 			exit(0);
