@@ -24,12 +24,11 @@
 #include <stdint.h>
 #include <inttypes.h>
 
-#include <linux/gfs_ondisk.h>
+#include "gfs_ondisk.h"
 #define __user
-#include <linux/gfs_ioctl.h>
+#include "gfs_ioctl.h"
 #include "osi_list.h"
-typedef uint64_t uint64;
-#include "iddev.h"
+#include "libgfs.h"
 
 #include "copyright.cf"
 
@@ -840,7 +839,7 @@ main(int argc, char *argv[])
 			break;
 		case 'V':
 			printf("%s %s (built %s %s)\n", argv[0],
-			       GFS_RELEASE_NAME, __DATE__, __TIME__);
+			       RELEASE_VERSION, __DATE__, __TIME__);
 			printf("%s\n", REDHAT_COPYRIGHT);
 			exit(0);
 		case 'h':
